# -*- coding: utf-8 -*-

from typing import Optional, List
from dataclasses import dataclass

import pandas as pd
import pandas_flavor as pf
from ._helpers import control_duplicates, control_value


@pf.register_dataframe_method
def difference(dataframe1, dataframe2) -> pd.DataFrame:
    """The set difference between dataframe1 (S) and dataframe2 (T), i.e. it returns those elements that are in dataframe1
    but not in dataframe2. Formally S - T = {s|s ∈ S and s ∉ T}.

    If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

    Does not alter the original DataFrame.

    Parameters
    ----------
    dataframe1 : pd.DataFrame\n
    dataframe2 : pd.DataFrame\n

    Returns
    -------
    pandas DataFrame\n
        The set difference between dataframe1 and dataframe2

    Raises
    ------
    ValueError\n
        Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
    """
    return SetOperations(dataframe1, dataframe2).difference()


@pf.register_dataframe_method
def symmetric_difference(
    dataframe1, dataframe2, dataframe_names: Optional[List[str]] = None
) -> pd.DataFrame:
    """The symmetric set difference between dataframe1 (S) and dataframe2 (T). It is the same as the union of the set difference between
    dataframe1 and dataframe2 and the set difference between dataframe2 and dataframe1. Formally A ⊕ B = (A - B) ∪ (B - A).

    If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

    Does not alter the original DataFrame.

    Parameters
    ----------
    dataframe1 : pd.DataFrame\n
    dataframe2 : pd.DataFrame\n
    dataframe_names : Optional[List[str]], default None\n
        The names given in the list is inserted in the returned dataframe in a column called 'original_dataframe'. The purpose
        of this column is to make it easier to track differences between the two dataframes.

    Returns
    -------
    pandas DataFrame\n
        The set symmetric difference between dataframe1 and dataframe2

    Raises
    ------
    ValueError\n
        Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
    ValueError\n
        Raises ValueError if the dataframe_names parameter is used and the length of the passed list is not 2.

    Example
    -------
    ```python
    import pandas as pd
    import neat_panda

    print(df1)

        country  continent  year  actual
    0   Sweden     Europe  2018       1
    1   Sweden     Europe  2019       2
    2  Denmark  Not known  2018       3

    print(df2)

        country  continent  year  actual
    0    Sweden     Europe  2018       1
    1   Denmark  Not known  2018       3
    2  Iceleand     Europe  2019       0

    df3 = df1.symmetric_difference(df2, dataframe_names=["df1", "df2"])
    print(df3)

    country       continent  year  actual original_dataframe
    0   Sweden       Europe  2019       2                df1
    1  Denmark    Not known  2018       3                df1
    2   Sweden       Europe  2012       2                df2
    3  Finland  Scandinavia  2018       3                df2
    ```
    """
    return SetOperations(dataframe1, dataframe2).symmetric_difference(
        dataframe_names=dataframe_names
    )


@pf.register_dataframe_method
def intersection(dataframe1, dataframe2) -> pd.DataFrame:
    """The set intersection between dataframe1 (S) and dataframe2 (T), i.e. it returns the elements that are both in dataframe1
    and dataframe2. Formally S ∩ T = {s|s ∈ S and s ∈ T}.

    If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

    Does not alter the original DataFrame.

    Parameters
    ----------
    dataframe1 : pd.DataFrame\n
    dataframe2 : pd.DataFrame\n

    Returns
    -------
    pandas DataFrame\n
        The set intersection between dataframe1 and dataframe2

    Raises
    ------
    ValueError\n
        Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
    """
    return SetOperations(dataframe1, dataframe2).intersection()


@pf.register_dataframe_method
def union(dataframe1, dataframe2) -> pd.DataFrame:
    """The set union between dataframe1 (S) and dataframe2 (T), i.e. it returns the elements that are both in dataframe1
    and dataframe2. Formally S ∩ T = {s|s ∈ S and s ∈ T}.

    If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

    Does not alter the original DataFrame.

    Syntactic sugar for the pandas dataframe append method.

    Parameters
    ----------
    dataframe1 : pd.DataFrame\n
    dataframe2 : pd.DataFrame\n

    Returns
    -------
    pandas DataFrame\n
        The set difference between dataframe1 and dataframe2

    Raises
    ------
    ValueError\n
        Raises ValueError if the columns in datframe1 and dataframe2 are not identical.

    Example
    -------
    ```python
    import panda as pd
    import neat_panda

    print(df1)

        country  continent  year  actual
    0    Sweden     Europe  2018       1
    1   Denmark  Not known  2018       3
    2  Iceleand     Europe  2019       0

    print(df2)

        country  continent  year  actual
    0    Sweden     Europe  2020       1
    1   Denmark  Not known  2020       3

    df3 = df1.union(df2)
    print(df3)

        country  continent  year  actual
    0    Sweden     Europe  2018       1
    1   Denmark  Not known  2018       3
    2  Iceleand     Europe  2019       0
    3    Sweden     Europe  2020       1
    4   Denmark  Not known  2020       3
    ```
    """
    return SetOperations(dataframe1, dataframe2).union()


@dataclass
class SetOperations:
    dataframe1: pd.DataFrame
    dataframe2: pd.DataFrame

    @control_value
    @control_duplicates
    def difference(self) -> pd.DataFrame:
        """The set difference between dataframe1 (S) and dataframe2 (T), i.e. it returns those elements that are in dataframe1
        but not in dataframe2. Formally S - T = {s|s ∈ S and s ∉ T}.

        If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

        Does not alter the original DataFrame.

        Parameters
        ----------
        dataframe1 : pd.DataFrame\n
        dataframe2 : pd.DataFrame\n

        The code is copied from the users WeNYoBen answer to the question "Python Pandas - Find difference between two data
        frames" at [StackOverflow](https://stackoverflow.com/questions/48647534/python-pandas-find-difference-between-two-data-frames)

        Returns
        -------
        pandas DataFrame\n
            The set difference between dataframe1 and dataframe2

        Raises
        ------
        ValueError\n
            Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
        """
        return pd.concat(
            [self.dataframe1, self.dataframe2, self.dataframe2]
        ).drop_duplicates(keep=False)

    @control_value
    @control_duplicates
    def symmetric_difference(
        self, dataframe_names: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """The symmetric set difference between dataframe1 (S) and dataframe2 (T). It is the same as the union of the set difference between
        dataframe1 and dataframe2 and the set difference between dataframe2 and dataframe1. Formally A ⊕ B = (A - B) ∪ (B - A).

        If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

        Does not alter the original DataFrame.

        Parameters
        ----------
        dataframe1 : pd.DataFrame\n
        dataframe2 : pd.DataFrame\n
        dataframe_names : Optional[List[str]], default None\n
            The names given in the list is inserted in the returned dataframe in a column called 'original_dataframe'. The purpose
            of this column is to make it easier to track differences between the two dataframes.

        Returns
        -------
        pandas DataFrame\n
            The set symmetric difference between dataframe1 and dataframe2

        Raises
        ------
        ValueError\n
            Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
        ValueError\n
            Raises ValueError if the dataframe_names parameter is used and the length of the passed list is not 2.
        """
        if dataframe_names and len(dataframe_names) != 2:
            raise ValueError("Only two dataframe names")
        df1 = self.difference()
        self._swap_dataframes()
        df2 = self.difference()
        if dataframe_names:
            df1["original_dataframe"] = dataframe_names[0]
            df2["original_dataframe"] = dataframe_names[1]
        return pd.concat([df1, df2]).reset_index(drop=True)

    def _swap_dataframes(self):
        temp = self.dataframe1.copy()
        self.dataframe1 = self.dataframe2.copy()
        self.dataframe2 = temp.copy()

    @control_value
    @control_duplicates
    def intersection(self) -> pd.DataFrame:
        """The set intersection between dataframe1 (S) and dataframe2 (T), i.e. it returns the elements that are both in dataframe1
        and dataframe2. Formally S ∩ T = {s|s ∈ S and s ∈ T}.

        If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

        Does not alter the original DataFrame.

        Parameters
        ----------
        dataframe1 : pd.DataFrame\n
        dataframe2 : pd.DataFrame\n

        Returns
        -------
        pandas DataFrame\n
            The set intersection between dataframe1 and dataframe2

        Raises
        ------
        ValueError\n
            Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
        """
        return self.dataframe1.merge(right=self.dataframe2, how="inner").reset_index(
            drop=True
        )

    @control_value
    @control_duplicates
    def union(self) -> pd.DataFrame:
        """The set union between dataframe1 (S) and dataframe2 (T), i.e. it returns the elements that are both in dataframe1
        and dataframe2. Formally S ∩ T = {s|s ∈ S and s ∈ T}.

        If duplicates exists in either dataframe they are dropped and a UserWarning is issued.

        Does not alter the original DataFrame.

        Syntactic sugar for the pandas dataframe append method.

        Parameters
        ----------
        dataframe1 : pd.DataFrame\n
        dataframe2 : pd.DataFrame\n

        Returns
        -------
        pandas DataFrame\n
            The set difference between dataframe1 and dataframe2

        Raises
        ------
        ValueError\n
            Raises ValueError if the columns in datframe1 and dataframe2 are not identical.
        """
        return self.dataframe1.append(self.dataframe2).reset_index(drop=True)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Top-level package for Neat Panda."""

__author__ = """Henric Sundberg"""
__email__ = "henric.sundberg@gmail.com"
__version__ = "0.9.7"

from ._tidy import spread, gather, flatten_pivot

# from ._caretaker import clean_column_names, _clean_column_names
from ._caretaker import clean_column_names, CleanColumnNames, clean_strings

from ._set_operations import (
    difference,
    intersection,
    symmetric_difference,
    union,
    SetOperations,
)

from ._helpers import _get_version_from_toml



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neat_panda']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.0.3', 'pandas_flavor>=0.1.2', 'toml>=0.10.0']

setup_kwargs = {
    'name': 'neat-panda',
    'version': '0.9.7',
    'description': '',
    'long_description': '\n# Neat Panda\n\n[![pypi](https://img.shields.io/pypi/v/neat_panda.svg)](https://pypi.python.org/pypi/neat_panda)\n[![Build Status](https://dev.azure.com/henricsundberg/neat_panda/_apis/build/status/htp84.neat_panda?branchName=master)](https://dev.azure.com/henricsundberg/neat_panda/_build/latest?definitionId=1&branchName=master)\n[![](https://img.shields.io/azure-devops/build/henricsundberg/neat_panda/1/master.svg)]()\n[![](https://img.shields.io/azure-devops/coverage/henricsundberg/neat_panda/1/master.svg)]()\n[![](https://img.shields.io/azure-devops/tests/henricsundberg/neat_panda/1/master.svg?passed_label=good&failed_label=bad&skipped_label=n%2Fa</code>)]()\n[![Supported](https://img.shields.io/pypi/pyversions/neat_panda.svg)](https://pypi.python.org/pypi/neat_panda)\n\n\nNeat Panda contains three main methods/functions, spread, gather and clean_columnames. The ideas for these methods are from the spread and gather functions in the R package [*tidyr*](https://tidyr.tidyverse.org/) and the make_clean_columns function in the R package [*janitor*](https://github.com/sfirke/janitor).\n\nThe spread function is syntactic sugar for the [*pandas*](https://pandas.pydata.org/pandas-docs/stable/) library method [*pivot*](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.pivot.html) and the gather method is syntactic sugar for the pandas method [*melt*](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.melt.html).\n\n\n## Features\n### clean_column_names\n```python\nimport neat_panda\n\nprint(df.columns.tolist())\n["Country    ", "Sub$region", "Actual"]\n\ndf = df.clean_column_names()\n\nprint(df.columns.tolist())\n["country", "sub_region", "actual"]\n\n```\n\n### spread\n#### R\n```R\nlibrary(tidyr)\nlibrary(dplyr)\nlibrary(gapminder)\n\ngapminder2 <- gapminder %>% select(country, continent, year, pop)\ngapminder3 <- gapminder2 %>% spread(key = year, value = pop)\nhead(gapminder3, n = 5)\n```\n#### Python\n```python\nimport neat_panda\nfrom gapminder import gapminder\n\ngapminder2 = gapminder[["country", "continent", "year", "pop"]]\ngapminder3 = gapminder2.spread(key="year", value="pop")\n\ngapminder3.head()\n```\n##### Output R\n```\n# A tibble: 5 x 14\n  country     continent   `1952`   `1957`   `1962`   `1967`   `1972`   `1977`   `1982`   `1987`   `1992`   `1997`   `2002`   `2007`\n  <fct>       <fct>        <int>    <int>    <int>    <int>    <int>    <int>    <int>    <int>    <int>    <int>    <int>    <int>\n1 Afghanistan Asia       8425333  9240934 10267083 11537966 13079460 14880372 12881816 13867957 16317921 22227415 25268405 31889923\n2 Albania     Europe     1282697  1476505  1728137  1984060  2263554  2509048  2780097  3075321  3326498  3428038  3508512  3600523\n3 Algeria     Africa     9279525 10270856 11000948 12760499 14760787 17152804 20033753 23254956 26298373 29072015 31287142 33333216\n4 Angola      Africa     4232095  4561361  4826015  5247469  5894858  6162675  7016384  7874230  8735988  9875024 10866106 12420476\n5 Argentina   Americas  17876956 19610538 21283783 22934225 24779799 26983828 29341374 31620918 33958947 36203463 38331121 40301927\n```\n##### Output Python\n```\n       country continent      1952      1957      1962      1967      1972      1977      1982      1987      1992      1997      2002      2007\n0  Afghanistan      Asia   8425333   9240934  10267083  11537966  13079460  14880372  12881816  13867957  16317921  22227415  25268405  31889923\n1      Albania    Europe   1282697   1476505   1728137   1984060   2263554   2509048   2780097   3075321   3326498   3428038   3508512   3600523\n2      Algeria    Africa   9279525  10270856  11000948  12760499  14760787  17152804  20033753  23254956  26298373  29072015  31287142  33333216\n3       Angola    Africa   4232095   4561361   4826015   5247469   5894858   6162675   7016384   7874230   8735988   9875024  10866106  12420476\n4    Argentina  Americas  17876956  19610538  21283783  22934225  24779799  26983828  29341374  31620918  33958947  36203463  38331121  40301927\n```\n\n\n### gather\n#### R\n```R\nlibrary(tidyr)\n\n# gapminder3 is obtained as above\ngapminder4 <- gather(gapminder3, key="year", "value"="pop", 3:14)\n# or\nyears <- c("1952", "1957", "1962", "1967", "1972", "1977", "1982", "1987", "1992", "1997", "2002", "2007")\ngapminder4 <- gather(gapminder3, key="year", "value"="pop", years)\n\nhead(gapminder4, n = 5)\n```\n#### Python\n```python\nimport neat_panda\n\n# gapminder3 is obtained as above\ngapminder4 = gapminder3.gather(key="year", value="pop", columns=range(2, 13))\n# or\ngapminder4 = gapminder3.gather(key="year", value="pop", columns=range(0, 2), invert_columns=True)\n# or\nyears = ["1952", "1957", "1962", "1967", "1972", "1977", "1982", "1987", "1992", "1997", "2002", "2007"]\ngapminder4 = gapminder3.gather(key="year", value="pop", columns=years)\n# or\ngapminder4 = gapminder3.gather(key="year", value="pop", columns=["country", "continent"], invert_columns=True)\n\ngapminder4.head()\n```\n##### Output R\n```\n# A tibble: 5 x 4\n  country     continent year       pop\n  <fct>       <fct>     <chr>    <int>\n1 Afghanistan Asia      1952   8425333\n2 Albania     Europe    1952   1282697\n3 Algeria     Africa    1952   9279525\n4 Angola      Africa    1952   4232095\n5 Argentina   Americas  1952  17876956\n```\n##### Output Python\n```\n       country continent  year       pop\n0  Afghanistan      Asia  1952   8425333\n1      Albania    Europe  1952   1282697\n2      Algeria    Africa  1952   9279525\n3       Angola    Africa  1952   4232095\n4    Argentina  Americas  1952  17876956\n\n```\n',
    'author': 'Henric Sundberg,',
    'author_email': 'henric.sundberg@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/htp84/neat_panda',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
